;
; stacktor-words.scm
; stacktor
;
; Useful words for stack manipulation and other actions.
; 
; Created by Mark Fredrickson on 2008-02-03.
; Copyright 2008. All rights reserved.
;

(require-for-syntax 'stacktor)
(use stacktor)
(use stacktor-base)
(require-for-syntax 'stack)
(use stack)

; =================
; = Helper macros =
; =================

(define-macro (arity-test stk n . body)
  (let ((stk-sym (gensym))
        (n-sym (gensym)))
    `(let ((,stk-sym ,stk) (,n-sym ,n)) 
      (if (> (- ,n-sym 1) (stack-depth ,stk-sym)) 
        (error "Insufficient stack depth")
        (begin ,@body)))))
  

; ================================================================================================
; = Stack Operations Words -- http://factorcode.org/responder/help/show-help?topic=shuffle-words =
; ================================================================================================
; Drops items from the stack
(define stk-drop (make-word (lambda (stk) (arity-test stk 1 (stack-pop! stk) stk))))
(define-word stk-2drop stk-drop stk-drop)
(define-word stk-3drop stk-drop stk-2drop)

; Drops non-top items
(define stk-nip (make-word (lambda (stk) 
  (arity-test stk 2 
    (let ((top (stack-pop! stk))) (begin (stack-pop! stk) (stack-push! stk top)))))))
(define stk-2nip (make-word (lambda (stk) 
  (arity-test stk 3 
    (let ((top (stack-pop! stk))) (begin (stack-cut! stk 0 1) (stack-push! stk top)))))))
    
; Duplicating items
(define stk-dup (make-word (lambda (stk)
  (arity-test stk 1
    (stack-push! stk (stack-peek stk 0))))))
(define stk-2dup (make-word (lambda (stk)
  (arity-test stk 2 ; because the stack is mutable, we peek at position 1 twice (though the value changes)
      (begin-with-stack stk (stack-peek stk 1) (stack-peek stk 1))))))
(define stk-3dup (make-word (lambda (stk)
  (arity-test stk 3 ; because the stack is mutable, we peek at position 1 twice (though the value changes)
      (begin-with-stack stk (stack-peek stk 2) (stack-peek stk 2) (stack-peek stk 2))))))
(define stk-dupd (make-word (lambda (stk)
  (arity-test stk 2
    (let ((top (stack-pop! stk))
          (next (stack-pop! stk)))
      (begin-with-stack stk next next top))))))
(define stk-over (make-word (lambda (stk) (arity-test stk 2 (begin-with-stack stk (stack-peek stk 1))))))
(define stk-pick (make-word (lambda (stk) (arity-test stk 3 (begin-with-stack stk (stack-peek stk 2))))))
(define stk-tuck (make-word (lambda (stk) (arity-test stk 2 (begin-with-stack stk stk-swap (stack-peek stk 1))))))

; Permuting Stack Elements
(define stk-swap (make-word (lambda (stk)
  (arity-test stk 2
    (let ((top (stack-pop! stk))
          (next (stack-pop! stk)))
      (begin-with-stack stk top next))))))
(define stk-swapd (make-word (lambda (stk) 
  (arity-test stk 3 (let ((top (stack-pop! stk))) (begin-with-stack stk stk-swap top))))))
(define stk-rot (make-word (lambda (stk)
  (arity-test stk 3 
    (let ((z (stack-pop! stk))
          (y (stack-pop! stk))
          (x (stack-pop! stk)))
      (begin-with-stack stk y z x))))))
(define stk--rot (make-word (lambda (stk)
  (arity-test stk 3
    (let ((z (stack-pop! stk))
          (y (stack-pop! stk))
          (x (stack-pop! stk)))
      (begin-with-stack stk z x y))))))      
(define stk-roll (make-word (lambda (stk)
  (arity-test stk 4
    (let ((t (stack-pop! stk))
          (z (stack-pop! stk))
          (y (stack-pop! stk))
          (x (stack-pop! stk)))
      (begin-with-stack stk y z t x))))))
(define stk--roll (make-word (lambda (stk)
  (arity-test stk 4
    (let ((t (stack-pop! stk))
          (z (stack-pop! stk))
          (y (stack-pop! stk))
          (x (stack-pop! stk)))
      (begin-with-stack stk t x y z))))))