;
; stacktor.scm
; stacktor: a practical Scheme stack DSL, inspired by Factor
; 
; Created by Mark Fredrickson on 2008-01-29.
; Copyright 2008. All rights reserved.
;

(use srfi-1)
(use stack)
(use stacktor-base)

;;; Helper function for stacktor procedure shadow
(define (force-arity sym . args) `(,sym (lambda (,@args . rst) (apply ,sym (append (list ,@args) rst)))))

;;; Shadowing basic functions with a macro
;;; This macro is unhygenic for a reason: to introduce local bindings for C_* functions that would
;;; otherwise fail arity checks.
(define-macro (stacktor-procedure-shadow . body)
  (let ((binary '(+ - / * < > = >= <= apply expt quotient))
        (unary '(string->number exact->inexact))
        (thunks '(argv)))
      `(let (,@(map (lambda (x) (force-arity x 'a 'b)) binary)
             ,@(map (lambda (x) (force-arity x 'a)) unary)
             ,@(map (lambda (x) (force-arity x)) thunks)) 
          ,@body)))
      
;;; The begin stack macro
(define-macro (begin-stack . body) `(begin-with-stack (make-stack) ,@body))
    
(define-macro (begin-with-stack stk . body)
  `(let ((stk-sym ,stk))
    (stacktor-procedure-shadow  
      (,@(fold (lambda (w lst) (list w lst)) 'stk-sym (map (lambda (x) `(stacktor-wrap ,x)) body))))))

;; word creation macro
(define-macro (word . words)
  `(make-word (lambda (stk) (begin-with-stack stk ,@words))))

;; named word creation
(define-macro (define-word name . words)
  `(define ,name (word ,@words)))
  
